"""S3 Op Class."""

import copy
import logging
import os
from typing import Sequence

from uniflow.node import Node
from uniflow.op.op import Op

logger = logging.getLogger(__name__)


class ExtractS3Op(Op):
    """Op to download a file from s3."""

    LOCAL_FILE_PATH = "/tmp/aws/s3"

    def __init__(self, name: str = "extract_s3_op") -> None:
        try:
            import boto3  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError("Please install boto3 to use S3Op.") from e

        self._s3_client = boto3.client("s3")
        super().__init__(name=name)

    def __call__(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run Model Op.

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        output_nodes = []
        for node in nodes:
            value_dict = copy.deepcopy(node.value_dict)
            # create local file path if not exists
            if os.path.exists(self.LOCAL_FILE_PATH) is False:
                os.makedirs(self.LOCAL_FILE_PATH)
            filename = os.path.join(self.LOCAL_FILE_PATH, value_dict["key"])
            logger.info("Downloading %s to %s", value_dict["key"], filename)
            self._s3_client.download_file(
                Bucket=value_dict["bucket"],
                Key=value_dict["key"],
                Filename=filename,
            )
            with open(
                filename,
                "r",
                encoding=value_dict.get("encoding", "utf-8"),
            ) as f:
                text = f.read()
            output_nodes.append(
                Node(
                    name=self.unique_name(),
                    value_dict={"text": text},
                    prev_nodes=[node],
                )
            )
        return output_nodes



"""LLM rater for tasks with a LLM model server."""

import re
from collections import Counter
from typing import Any, Dict, List

from uniflow.op.model.constants import (
    AVERAGE_SCORE,
    MAJORITY_VOTE,
    RESPONSE,
    SAMPLES,
    SCORES,
    VOTES,
)
from uniflow.op.model.llm_processor import JsonFormattedDataProcessor, LLMDataProcessor
from uniflow.op.prompt import PromptTemplate


class LLMRater(LLMDataProcessor):
    """LLM Rater"""

    def __init__(
        self,
        prompt_template: PromptTemplate,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """LLM Rater Constructor.

        Args:
            prompt_template (PromptTemplate): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__(prompt_template, model_config)
        pattern = r"^[^A-Za-z]+|[^A-Za-z]+$"
        self._label2score = {
            re.sub(pattern, "", k).lower().lower(): float(v)
            for k, v in label2score.items()
        }
        self._score2label = {v: k for k, v in self._label2score.items()}

    def _deserialize(self, data: List[str]) -> List[Dict[str, Any]]:
        """Deserialize data.

        Args:
            data (List[str]): Data to deserialize.

        Returns:
            List[Dict[str, Any]]: Deserialized data.
        """

        def _extract_label(text):
            # Create a regular expression pattern from the dictionary keys
            # The pattern will look like '(yes|no)', case-insensitive
            pattern = (
                r"\b("
                + "|".join(  # noqa: W503
                    re.escape(key) for key in self._label2score.keys()
                )  # noqa: W503
                + r")\b"  # noqa: W503
            )

            matches = list(re.finditer(pattern, text, re.IGNORECASE))
            if matches:
                # Return the last matched group (Yes or No), preserving the original case
                return (
                    matches[-1].group(1).lower()
                )  # Convert to lowercase to match dictionary keys
            else:
                # Return None if no match is found
                return None

        data = super()._deserialize(data)
        response = data[RESPONSE]
        reformatted_responses = []
        for i in range(0, len(response), self._num_samples):
            samples = response[i : i + self._num_samples]  # noqa: E203

            labels = [_extract_label(d) for d in samples]
            scores = []
            for label in labels:
                if label is not None:
                    scores.append(self._label2score[label])
            majority_vote = Counter(labels).most_common(1)[0][0]
            mean_score = sum(scores) / len(scores) if len(scores) > 0 else None

            reformatted_responses.append(
                {
                    SAMPLES: samples,
                    MAJORITY_VOTE: majority_vote,
                    AVERAGE_SCORE: mean_score,
                    VOTES: labels,
                    SCORES: scores,
                }
            )
        data[RESPONSE] = reformatted_responses

        return data


class OpenAIJsonFormattedLLMRater(JsonFormattedDataProcessor):
    """OpenAI formatted LLM Rater"""

    def __init__(
        self,
        prompt_template: PromptTemplate,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """Json Formatted LLM Rater Constructor.

        Args:
            prompt_template (PromptTemplate): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__(prompt_template, model_config)
        self._pattern = r"^[^A-Za-z]+|[^A-Za-z]+$"
        self._label2score = {
            re.sub(self._pattern, "", k).lower(): float(v)
            for k, v in label2score.items()
        }
        self._score2label = {v: k for k, v in self._label2score.items()}
        self._rater_key = None
        if prompt_template.few_shot_prompt:
            example_keys = list(prompt_template.few_shot_prompt[0].dict().keys())
            self._rater_key = example_keys[-1]

    def _deserialize(self, data: List[str]) -> List[Dict[str, Any]]:
        """Deserialize data.

        Args:
            data (List[str]): Data to deserialize.

        Returns:
            List[Dict[str, Any]]: Deserialized data.
        """
        data = super()._deserialize(data)
        response = data[RESPONSE]
        reformatted_responses = []

        for i in range(0, len(response), self._num_samples):
            samples = response[i : i + self._num_samples]  # noqa: E203
            if self._rater_key:
                labels = [
                    re.sub(self._pattern, "", r[self._rater_key]).lower()
                    if self._rater_key in r
                    else None
                    for r in samples
                ]
            else:
                # If the rater key is not specified, use the last key in the response
                # as the rater key for the first response.
                self._rater_key = list(response[0].keys())[-1]
                labels = [
                    re.sub(self._pattern, "", r[self._rater_key]).lower()
                    for r in samples
                ]
            scores = []
            for label in labels:
                if label is not None and label in self._label2score:
                    scores.append(self._label2score[label])
            majority_vote = Counter(labels).most_common(1)[0][0]
            mean_score = sum(scores) / len(scores) if len(scores) > 0 else None
            reformatted_responses.append(
                {
                    SAMPLES: samples,
                    MAJORITY_VOTE: majority_vote,
                    AVERAGE_SCORE: mean_score,
                    VOTES: labels,
                    SCORES: scores,
                }
            )
        data[RESPONSE] = reformatted_responses
        return data


class HuggingfaceJsonFormattedLLMRater(LLMDataProcessor):
    """Huggingface formatted Json Formatted LLM Rater"""

    def __init__(
        self,
        prompt_template: PromptTemplate,
        model_config: Dict[str, Any],
        label2score: Dict[str, float],
    ) -> None:
        """Json Formatted LLM Rater Constructor.

        Args:
            prompt_template (PromptTemplate): Guided prompt template.
            model_config (Dict[str, Any]): Model config.
            label2score (Dict[str, float]): String to score mapping.
        """
        super().__init__(prompt_template, model_config)
        self._pattern = r"^[^A-Za-z]+|[^A-Za-z]+$"
        self._label2score = {
            re.sub(self._pattern, "", k).lower(): float(v)
            for k, v in label2score.items()
        }
        self._score2label = {v: k for k, v in self._label2score.items()}
        self._rater_key = None
        if prompt_template.few_shot_prompt:
            example_keys = list(prompt_template.few_shot_prompt[0].dict().keys())
            self._rater_key = example_keys[-1]

    def _deserialize(self, data: List[str]) -> List[Dict[str, Any]]:
        """Deserialize data.

        Args:
            data (List[str]): Data to deserialize.

        Returns:
            List[Dict[str, Any]]: Deserialized data.
        """
        data = super()._deserialize(data)
        response = data[RESPONSE]
        reformatted_responses = []

        for i in range(0, len(response), self._num_samples):
            samples = response[i : i + self._num_samples]  # noqa: E203
            if self._rater_key:
                labels = [
                    re.sub(self._pattern, "", r[self._rater_key]).lower()
                    if self._rater_key in r
                    else None
                    for r in samples
                ]
            else:
                # If the rater key is not specified, use the last key in the response
                # as the rater key for the first response.
                self._rater_key = list(response[0].keys())[-1]
                labels = [
                    re.sub(self._pattern, "", r[self._rater_key]).lower()
                    for r in samples
                ]
            scores = []
            for label in labels:
                if label is not None and label in self._label2score:
                    scores.append(self._label2score[label])
            majority_vote = Counter(labels).most_common(1)[0][0]
            mean_score = sum(scores) / len(scores) if len(scores) > 0 else None
            reformatted_responses.append(
                {
                    SAMPLES: samples,
                    MAJORITY_VOTE: majority_vote,
                    AVERAGE_SCORE: mean_score,
                    VOTES: labels,
                    SCORES: scores,
                }
            )
        data[RESPONSE] = reformatted_responses
        return data



import hashlib
import os
import zipfile
from typing import Dict, List, Union

import requests
import torch
from tqdm import tqdm
from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

try:
    from transformers_neuronx import constants
    from transformers_neuronx.config import NeuronConfig
    from transformers_neuronx.generation_utils import HuggingFaceGenerationModelAdapter
    from transformers_neuronx.mistral.model import MistralForSampling
    from transformers_neuronx.module import save_pretrained_split
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Please install transformers_neuronx."
        "You can use `pip install transformers-neuronx --extra-index-url=https://pip.repos.neuron.amazonaws.com` to install it."
    ) from exc


class Neuron:
    """
    Neuron class.
    """

    neuron_model_config: Dict[str, Dict[int, Dict[int, List[Union[str, str]]]]] = {
        "mistralai/Mistral-7B-Instruct-v0.2": {
            1: {
                8192: [
                    "https://cambioml-neuron-cache.s3.us-west-2.amazonaws.com/transformers_neuronx/Mistral-7B-Instruct-v0.2_tp2-bz1-8192.zip",
                    "367aa7db73c1f82b5add0ad5c97d7a77",
                ]
            },
            2: {
                4096: [
                    "https://cambioml-neuron-cache.s3.us-west-2.amazonaws.com/transformers_neuronx/Mistral-7B-Instruct-v0.2_tp2-bz2-4096.zip",
                    "f2f07a6d6b05e39b7e9d94cd6c95fb96",
                ]
            },
            4: {
                2048: [
                    "https://cambioml-neuron-cache.s3.us-west-2.amazonaws.com/transformers_neuronx/Mistral-7B-Instruct-v0.2_tp2-bz4-2048.zip",
                    "1c62a06d6f25c2241b72b95e00ededfb",
                ]
            },
        }
    }
    instance_type_dict: Dict[str, int] = {
        "inf2.xlarge": 2,
        "inf2.8xlarge": 2,
        "inf2.24xlarge": 12,
        "inf2.48xlarge": 24,
    }

    @staticmethod
    def get_instance_type() -> str:
        """
        Get the instance type of the current instance.

        Returns:
            str: The instance type of the current AWS instance.

        Raises:
            Exception: If there is an error accessing the metadata service.

        References:
            - AWS EC2 User Guide: https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
        """
        url = "http://169.254.169.254/latest/meta-data/instance-type"
        try:
            response = requests.get(url, timeout=1)
            if response.status_code == 200:
                return response.text
            else:
                return "Error: Unable to access metadata service"
        except Exception as e:
            return f"Error: {e}"

    @staticmethod
    def verify_md5(file_path: str, expected_md5: str) -> bool:
        """
        Verify the MD5 of a file against an expected MD5 hash.

        Args:
            file_path (str): Path to the file to be checked.
            expected_md5 (str): Expected MD5 hash for the file.

        Returns:
            bool: True if MD5 matches, False otherwise.
        """
        md5_hash = hashlib.md5()
        with open(file_path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5_hash.update(chunk)
        file_md5 = md5_hash.hexdigest()
        return file_md5 == expected_md5

    @staticmethod
    def download_and_unzip(url: str, expected_md5: str) -> str:
        """
        Download a file from a URL and unzip it.

        Args:
            url (str): The URL of the file to download.
            expected_md5 (str): The expected MD5 hash of the file.

        Returns:
            str: The path to the unzipped file.

        Raises:
            ValueError: If the MD5 verification fails.
        """
        cache_dir = os.path.join(os.path.expanduser("~"), ".cache/uniflow/")
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        zip_filename = os.path.join(cache_dir, f"{expected_md5}.zip")
        if os.path.exists(os.path.join(cache_dir, f"{expected_md5}")):
            return os.path.join(cache_dir, f"{expected_md5}")
        response = requests.get(url, stream=True)
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024 * 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

        with open(zip_filename, "wb") as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if not Neuron.verify_md5(zip_filename, expected_md5):
            raise ValueError("MD5 verification failed")
        print(zip_filename, os.path.join(cache_dir, f"{expected_md5}"))
        with zipfile.ZipFile(zip_filename, "r") as zip_ref:
            zip_ref.extractall(os.path.join(cache_dir, f"{expected_md5}"))
        os.remove(zip_filename)
        return os.path.join(cache_dir, f"{expected_md5}")

    @staticmethod
    def split_neuron_model(model_name: str) -> str:
        """
        Split weights

        Args:
            model_name (str): The name of the model to split weights from.

        Returns:
            str: The directory path where the split weights are saved.
        """

        cache_dir = os.path.join(
            os.path.expanduser("~"),
            ".cache/uniflow/",
            model_name.split("/")[1] + "-split",
        )
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        model_cpu = AutoModelForCausalLM.from_pretrained(model_name)
        save_pretrained_split(model_cpu, cache_dir)
        return cache_dir

    @staticmethod
    def get_neuron_model(model_name: str, batch_size: int):
        """
        Get the neuron model for the given model name and batch size.

        Args:
            model_name (str): The name of the model.
            batch_size (int): The batch size.

        Returns:
            The neuron model.
        """
        instance_type = Neuron.get_instance_type()
        assert "inf2" in instance_type, ValueError(
            "Neuron model can only run on the AWS EC2 inf2 instance series."
        )
        tp_degree = Neuron.instance_type_dict[instance_type]
        assert model_name in Neuron.neuron_model_config, ValueError(
            f"Current Neuron only support {list(Neuron.neuron_model_config.keys())}"
        )
        assert batch_size in Neuron.neuron_model_config[model_name], ValueError(
            f"{model_name} only support batch size {list(Neuron.neuron_model_config[model_name].keys())} in {instance_type}"
        )
        n_positions_list = list(
            Neuron.neuron_model_config[model_name][batch_size].keys()
        )
        n_positions_list.sort()
        n_positions = max(Neuron.neuron_model_config[model_name][batch_size])
        neuron_config = NeuronConfig(
            grouped_query_attention=constants.GQA.SHARD_OVER_HEADS
        )
        cache_dir = os.path.join(
            os.path.expanduser("~"),
            ".cache/uniflow/",
            model_name.split("/")[1] + "-split",
        )
        if os.path.exists(cache_dir):
            print(
                f"Model chunks found. If you encounter errors, please delete the cache folder `{cache_dir}`"
            )
            model_split = cache_dir
        else:
            print("Spliting model, need to wait for a while...")
            model_split = Neuron.split_neuron_model(model_name)

        model_neuron = MistralForSampling.from_pretrained(
            model_split,
            batch_size=batch_size,
            tp_degree=tp_degree,
            n_positions=n_positions,
            amp="bf16",
            neuron_config=neuron_config,
        )
        if model_neuron.config.window_size is None:
            model_neuron.config.window_size = 4096
        url, md5 = Neuron.neuron_model_config[model_name][batch_size][n_positions]
        neuron_cache = Neuron.download_and_unzip(url, md5)
        print("loading neuron model from cache, need to wait for a while...")
        model_neuron.load(neuron_cache)
        model_neuron.to_neuron()
        model_neuron.eval()
        model_config = AutoConfig.from_pretrained(model_name)
        model = HuggingFaceGenerationModelAdapter(model_config, model_neuron)
        model.reset_generation()
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        tokenizer.pad_token_id = tokenizer.eos_token_id
        return model, tokenizer

    @staticmethod
    def batch_list(lst: List, batch_size: int) -> List[List]:
        """
        Split a list into batches of a specified size.

        Args:
            lst (list): The list to be split into batches.
            batch_size (int): The size of each batch.

        Returns:
            list: A list of batches, where each batch contains elements from the original list.
        """
        batches = []
        for i in range(0, len(lst), batch_size):
            batch = lst[i : i + batch_size]
            while len(batch) < batch_size:
                batch.append(lst[-1])
            batches.append(batch)
        return batches

    @staticmethod
    def neuron_infer(text_list: List[str], model, tokenizer) -> List[Dict[str, str]]:
        """
        Run neuron inference on a list of texts.

        Args:
            text_list (list): A list of texts to perform inference on.
            model: The neuron model used for inference.
            tokenizer: The tokenizer used to preprocess the texts.

        Returns:
            list: A list of dictionaries containing the generated text for each input text.
        """
        batches = Neuron.batch_list(text_list, 4)
        results = []
        for batch in batches:
            encoded_input = tokenizer(
                ["[INST]" + text + "[/INST]" for text in batch],
                return_tensors="pt",
                padding=True,
            )
            with torch.inference_mode():
                sample_output = model.generate(
                    input_ids=encoded_input.input_ids,
                    attention_mask=encoded_input.attention_mask,
                    do_sample=True,
                    max_length=1024,
                    eos_token_id=tokenizer.eos_token_id,
                    pad_token_id=tokenizer.pad_token_id,
                    temperature=0.7,
                )

            for sample_idx, tok in enumerate(sample_output):
                start = len(encoded_input.input_ids[sample_idx])
                tok = tok[start:]
                end = list(tok).index(tokenizer.eos_token_id)
                results.append(
                    [
                        {
                            "generated_text": (
                                batch[sample_idx] + tokenizer.decode(tok[:end])
                            )
                        }
                    ]
                )
        results = results[: len(text_list)]
        return results



"""All Uniflow Clients"""

from dataclasses import asdict
from typing import Any, List, Mapping

from uniflow.flow.config import ExtractConfig, RaterConfig, TransformConfig
from uniflow.flow.server import ExtractServer, RaterServer, TransformServer


class ExtractClient:
    """Uniflow Extract Client"""

    def __init__(self, config: ExtractConfig) -> None:
        """Client constructor

        Args:
            config (Config): Config for the flow

        """
        self._config = config
        self._server = ExtractServer(asdict(self._config))

    def run(self, input_list: List[Mapping[str, Any]]) -> List[Mapping[str, Any]]:
        """
        Run the flow

        Args:
            input_list (List[Mapping[str, Any]]): List of inputs to the flow

        Returns:
            List[Mapping[str, Any]]: List of outputs from the flow
        """
        # convert config to dict for future remote calls
        output = self._server.run(input_list)
        return output

    def async_run(self) -> None:
        """Run the flow asynchronously"""
        # TODO: Implement async client
        print("Client running async")


class TransformClient:
    """Uniflow Transform Client"""

    def __init__(self, config: TransformConfig) -> None:
        """Client constructor

        Args:
            config (Config): Config for the flow

        """
        self._config = config
        self._server = TransformServer(asdict(self._config))

    def run(self, input_list: List[Mapping[str, Any]]) -> List[Mapping[str, Any]]:
        """
        Run the flow

        Args:
            input_list (List[Mapping[str, Any]]): List of inputs to the flow

        Returns:
            List[Mapping[str, Any]]: List of outputs from the flow
        """
        # convert config to dict for future remote calls
        output = self._server.run(input_list)
        return output

    def async_run(self) -> None:
        """Run the flow asynchronously"""
        # TODO: Implement async client
        print("Client running async")


class RaterClient:
    """Uniflow Rater Client"""

    def __init__(self, config: RaterConfig) -> None:
        """Client constructor

        Args:
            config (Config): Config for the flow

        """
        self._config = config
        self._server = RaterServer(asdict(self._config))

    def run(self, input_list: List[Mapping[str, Any]]) -> List[Mapping[str, Any]]:
        """
        Run the flow

        Args:
            input_list (List[Mapping[str, Any]]): List of inputs to the flow

        Returns:
            List[Mapping[str, Any]]: List of outputs from the flow
        """
        # convert config to dict for future remote calls
        output = self._server.run(input_list)
        return output

    def async_run(self) -> None:
        """Run the flow asynchronously"""
        # TODO: Implement async client
        print("Client running async")



"""Layout Module."""
from typing import List

import cv2
import numpy as np
import onnxruntime
from huggingface_hub import hf_hub_download


class Layout:
    """
    Layout class.
    """

    @staticmethod
    def preprocess(
        img: np.ndarray, input_size: tuple, swap: tuple = (2, 0, 1)
    ) -> tuple[np.ndarray, float]:
        """
        Preprocesses the input image by resizing and padding it.

        Args:
            img (numpy.ndarray): The input image.
            input_size (tuple): The desired size of the input image.
            swap (tuple, optional): The order of dimensions to swap. Defaults to (2, 0, 1).

        Returns:
            tuple[numpy.ndarray, float]: The preprocessed image and the scaling factor applied to the image.
        """
        if len(img.shape) == 3:
            padded_img = (
                np.ones((input_size[0], input_size[1], 3), dtype=np.uint8) * 114
            )
        else:
            padded_img = np.ones(input_size, dtype=np.uint8) * 114

        r = min(input_size[0] / img.shape[0], input_size[1] / img.shape[1])
        resized_img = cv2.resize(
            img,
            (int(img.shape[1] * r), int(img.shape[0] * r)),
            interpolation=cv2.INTER_LINEAR,
        ).astype(np.uint8)
        padded_img[: int(img.shape[0] * r), : int(img.shape[1] * r)] = resized_img

        padded_img = padded_img.transpose(swap)
        padded_img = np.ascontiguousarray(padded_img, dtype=np.float32)
        return padded_img, r

    @staticmethod
    def nms(boxes: np.ndarray, scores: np.ndarray, nms_thr: float) -> List[int]:
        """Single class NMS implemented in Numpy."""
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]

        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[::-1]

        keep = []
        while order.size > 0:
            i = order[0]
            keep.append(i)
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])

            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (areas[i] + areas[order[1:]] - inter)

            inds = np.where(ovr <= nms_thr)[0]
            order = order[inds + 1]

        return keep

    @staticmethod
    def multiclass_nms(
        boxes: np.ndarray,
        scores: np.ndarray,
        nms_thr: float,
        score_thr: float,
        class_agnostic: bool = True,
    ) -> np.ndarray:
        """Multiclass NMS implemented in Numpy"""
        if class_agnostic:
            nms_method = Layout.multiclass_nms_class_agnostic
        else:
            nms_method = Layout.multiclass_nms_class_aware
        return nms_method(boxes, scores, nms_thr, score_thr)

    @staticmethod
    def multiclass_nms_class_aware(
        boxes: np.ndarray, scores: np.ndarray, nms_thr: float, score_thr: float
    ) -> np.ndarray:
        """Multiclass NMS implemented in Numpy. Class-aware version."""
        final_dets = []
        num_classes = scores.shape[1]
        for cls_ind in range(num_classes):
            cls_scores = scores[:, cls_ind]
            valid_score_mask = cls_scores > score_thr
            if valid_score_mask.sum() == 0:
                continue
            else:
                valid_scores = cls_scores[valid_score_mask]
                valid_boxes = boxes[valid_score_mask]
                keep = Layout.nms(valid_boxes, valid_scores, nms_thr)
                if len(keep) > 0:
                    cls_inds = np.ones((len(keep), 1)) * cls_ind
                    dets = np.concatenate(
                        [valid_boxes[keep], valid_scores[keep, None], cls_inds], 1
                    )
                    final_dets.append(dets)
        if len(final_dets) == 0:
            return None
        return np.concatenate(final_dets, 0)

    @staticmethod
    def multiclass_nms_class_agnostic(
        boxes: np.ndarray, scores: np.ndarray, nms_thr: float, score_thr: float
    ) -> np.ndarray:
        """Multiclass NMS implemented in Numpy. Class-agnostic version."""
        cls_inds = scores.argmax(1)
        cls_scores = scores[np.arange(len(cls_inds)), cls_inds]

        valid_score_mask = cls_scores > score_thr
        if valid_score_mask.sum() == 0:
            return None
        valid_scores = cls_scores[valid_score_mask]
        valid_boxes = boxes[valid_score_mask]
        valid_cls_inds = cls_inds[valid_score_mask]
        keep = Layout.nms(valid_boxes, valid_scores, nms_thr)
        if keep:
            dets = np.concatenate(
                [
                    valid_boxes[keep],
                    valid_scores[keep, None],
                    valid_cls_inds[keep, None],
                ],
                1,
            )
        return dets

    @staticmethod
    def postprocess(
        outputs: np.ndarray, img_size: tuple, p6: bool = False
    ) -> np.ndarray:
        """
        Postprocesses the outputs of a model by applying grid and stride calculations.

        Args:
            outputs (numpy.ndarray): The output predictions from the model.
            img_size (tuple): The size of the input image.
            p6 (bool, optional): Whether to include the stride 64 grid. Defaults to False.

        Returns:
            numpy.ndarray: The postprocessed outputs.

        """
        grids = []
        expanded_strides = []

        if not p6:
            strides = [8, 16, 32]
        else:
            strides = [8, 16, 32, 64]

        hsizes = [img_size[0] // stride for stride in strides]
        wsizes = [img_size[1] // stride for stride in strides]

        for hsize, wsize, stride in zip(hsizes, wsizes, strides):
            xv, yv = np.meshgrid(np.arange(wsize), np.arange(hsize))
            grid = np.stack((xv, yv), 2).reshape(1, -1, 2)
            grids.append(grid)
            shape = grid.shape[:2]
            expanded_strides.append(np.full((*shape, 1), stride))

        grids = np.concatenate(grids, 1)
        expanded_strides = np.concatenate(expanded_strides, 1)
        outputs[..., :2] = (outputs[..., :2] + grids) * expanded_strides
        outputs[..., 2:4] = np.exp(outputs[..., 2:4]) * expanded_strides

        return outputs


class XYCut:
    """
    XYCut class.
    """

    @staticmethod
    def projection_by_bboxes(boxes: np.array, axis: int) -> np.ndarray:
        """
        Get the projection histogram by a set of bounding boxes, and output it in per-pixel form.

        Args:
            boxes: [N, 4]
            axis: 0 for horizontal projection along the x-axis, 1 for vertical projection along the y-axis

        Returns:
            1D projection histogram with a length equal to the maximum value of the coordinate in the projection direction
            (We don't need the actual image size, as we are only interested in finding the gaps between text boxes)
        """
        assert axis in [0, 1]
        length = np.max(boxes[:, axis::2])
        res = np.zeros(length, dtype=int)
        for start, end in boxes[:, axis::2]:
            res[start:end] += 1
        return res

    @staticmethod
    def split_projection_profile(
        arr_values: np.array, min_value: float, min_gap: float
    ):
        """Split projection profile:
        from: https://dothinking.github.io/2021-06-19-%E9%80%92%E5%BD%92%E6%8A%95%E5%BD%B1%E5%88%86%E5%89%B2%E7%AE%97%E6%B3%95/#:~:text=%E9%80%92%E5%BD%92%E6%8A%95%E5%BD%B1%E5%88%86%E5%89%B2%EF%BC%88Recursive%20XY,%EF%BC%8C%E5%8F%AF%E4%BB%A5%E5%88%92%E5%88%86%E6%AE%B5%E8%90%BD%E3%80%81%E8%A1%8C%E3%80%82


        ```
                                ┌──┐
            arr_values           │  │       ┌─┐───
                ┌──┐             │  │       │ │ |
                │  │             │  │ ┌───┐ │ │min_value
                │  │<- min_gap ->│  │ │   │ │ │ |
            ────┴──┴─────────────┴──┴─┴───┴─┴─┴─┴───
            0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
        ```

        Args:
            arr_values (np.array): 1-d array representing the projection profile.
            min_value (float): Ignore the profile if `arr_value` is less than `min_value`.
            min_gap (float): Ignore the gap if less than this value.

        Returns:
            tuple: Start indexes and end indexes of split groups.
        """
        # all indexes with projection height exceeding the threshold
        arr_index = np.where(arr_values > min_value)[0]
        if len(arr_index) == 0:
            return

        # find zero intervals between adjacent projections
        # |  |                    ||
        # ||||<- zero-interval -> |||||
        arr_diff = arr_index[1:] - arr_index[0:-1]
        arr_diff_index = np.where(arr_diff > min_gap)[0]
        arr_zero_intvl_start = arr_index[arr_diff_index]
        arr_zero_intvl_end = arr_index[arr_diff_index + 1]

        # convert to index of projection range:
        # the start index of zero interval is the end index of projection
        arr_start = np.insert(arr_zero_intvl_end, 0, arr_index[0])
        arr_end = np.append(arr_zero_intvl_start, arr_index[-1])
        arr_end += 1  # end index will be excluded as index slice

        return arr_start, arr_end

    @staticmethod
    def recursive_xy_cut(boxes: np.ndarray, indices: List[int], res: List[int]):
        """
        from https://github.com/Sanster/xy-cut/blob/main/xycut.py
        Args:
            boxes: (N, 4)
            indices: Represents the indices of the boxes in the original data during the recursive process.
            res: Stores the output results.

        """
        assert len(boxes) == len(indices)

        _indices = boxes[:, 1].argsort()
        y_sorted_boxes = boxes[_indices]
        y_sorted_indices = indices[_indices]

        # debug_vis(y_sorted_boxes, y_sorted_indices)

        y_projection = XYCut.projection_by_bboxes(boxes=y_sorted_boxes, axis=1)
        pos_y = XYCut.split_projection_profile(y_projection, 0, 1)
        if not pos_y:
            return

        arr_y0, arr_y1 = pos_y
        for r0, r1 in zip(arr_y0, arr_y1):
            _indices = (r0 <= y_sorted_boxes[:, 1]) & (y_sorted_boxes[:, 1] < r1)

            y_sorted_boxes_chunk = y_sorted_boxes[_indices]
            y_sorted_indices_chunk = y_sorted_indices[_indices]

            _indices = y_sorted_boxes_chunk[:, 0].argsort()
            x_sorted_boxes_chunk = y_sorted_boxes_chunk[_indices]
            x_sorted_indices_chunk = y_sorted_indices_chunk[_indices]

            x_projection = XYCut.projection_by_bboxes(
                boxes=x_sorted_boxes_chunk, axis=0
            )
            pos_x = XYCut.split_projection_profile(x_projection, 0, 1)
            if not pos_x:
                continue

            arr_x0, arr_x1 = pos_x
            if len(arr_x0) == 1:
                res.extend(x_sorted_indices_chunk)
                continue

            for c0, c1 in zip(arr_x0, arr_x1):
                _indices = (c0 <= x_sorted_boxes_chunk[:, 0]) & (
                    x_sorted_boxes_chunk[:, 0] < c1
                )
                XYCut.recursive_xy_cut(
                    x_sorted_boxes_chunk[_indices],
                    x_sorted_indices_chunk[_indices],
                    res,
                )


class LayoutPredictor:
    """
    Class for predicting layout of an image.

    Attributes:
        ort_session (onnxruntime.InferenceSession): Inference session for running the model.
        categorys (dict): Mapping of category index to category name.
    """

    H = 1024
    W = 768

    def __init__(self, model_name: str, model_file: str):
        """
        Initializes the LayoutPredictor class.
        Downloads the model specified by `model_name` and `model_file` from the Hugging Face Model Hub.
        Creates an inference session using the downloaded model.
        """
        model_path = hf_hub_download(model_name, model_file)
        self.ort_session = onnxruntime.InferenceSession(
            model_path, providers=["CUDAExecutionProvider", "CPUExecutionProvider"]
        )
        self.categorys = {
            0: "Caption",
            1: "Footnote",
            2: "Formula",
            3: "List-item",
            4: "Page-footer",
            5: "Page-header",
            6: "Picture",
            7: "Section-header",
            8: "Table",
            9: "Text",
            10: "Title",
        }

    def __call__(self, img):
        """
        Predicts the layout of the input image.

        Args:
            img (numpy.ndarray): Input image.

        Returns:
            list: List of dictionaries containing the predicted layout information.
                Each dictionary contains the type of the element and its bounding box coordinates.
        """
        image, ratio = Layout.preprocess(img, (self.H, self.W))
        res = self.ort_session.run(["output"], {"images": image[np.newaxis, :]})[0]
        predictions = Layout.postprocess(res, (self.H, self.W), p6=False)[0]
        boxes = predictions[:, :4]
        scores = predictions[:, 4, None] * predictions[:, 5:]
        boxes_xyxy = np.ones_like(boxes)
        boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.0
        boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.0
        boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.0
        boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.0
        boxes_xyxy /= ratio
        dets = Layout.multiclass_nms(boxes_xyxy, scores, nms_thr=0.15, score_thr=0.3)
        result = []
        if dets is None:
            return result
        scores = dets[:, 4]
        final_cls_inds = dets[:, 5]
        final_boxes = dets[:, :4].astype("int")

        for box_idx, box in enumerate(final_boxes):
            result.append(
                {"type": self.categorys[int(final_cls_inds[box_idx])], "bbox": box}
            )
        return result



"""Extract PDF Flow."""

from typing import Any, Dict, Sequence

from uniflow.constants import EXTRACT
from uniflow.flow.flow import Flow
from uniflow.node import Node
from uniflow.op.extract.load.pdf_op import ExtractPDFOp, ProcessPDFOp
from uniflow.op.extract.split.constants import PARAGRAPH_SPLITTER
from uniflow.op.extract.split.splitter_factory import SplitterOpsFactory
from uniflow.op.model.llm_preprocessor import LLMDataPreprocessor


class ExtractPDFFlow(Flow):
    """Extract PDF Flow Class."""

    TAG = EXTRACT

    def __init__(
        self,
        model_config: Dict[str, Any],
        splitter: str = PARAGRAPH_SPLITTER,
    ) -> None:
        """Extract PDF Flow Constructor.

        Args:
            model_config (Dict[str, Any]): Model config.
            splitter (str): Splitter to use. Defaults to "".
        """
        super().__init__()
        self._extract_pdf_op = ExtractPDFOp(
            name="extract_pdf_op",
            model=LLMDataPreprocessor(
                model_config=model_config,
            ),
        )
        self._process_pdf_op = ProcessPDFOp(name="process_pdf_op")
        self._split_op = SplitterOpsFactory.get(splitter)

    def run(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run Model Flow.

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        nodes = self._extract_pdf_op(nodes)
        nodes = self._process_pdf_op(nodes)
        nodes = self._split_op(nodes)
        return nodes



"""ops utils"""
import logging
import os

OPS_SCOPE = []

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_op_scope_name(name: str) -> str:
    """Get op scope name.

    Args:
        name (str): Name of the op.

    Returns:
        str: Op scope name.
    """
    return "/".join(OPS_SCOPE + [name])


def check_path_exists(save_path: str) -> None:
    """Check if path exists, if not create it.

    Args:
        save_path (str): Path to check.

    Returns:
        None: Nothing.
    """
    if not os.path.exists(save_path):
        try:
            # Create the directory and any missing parent directories
            os.makedirs(save_path)
            logger.info("Directory %s created successfully.", save_path)
        except OSError as e:
            logger.info("Error creating directory %s: %s", save_path, e)
    else:
        logger.info("Directory %s already exists.", save_path)



"""Model __init__ Module."""

# this register all possible model server into ModelServerFactory through
# ModelServerFactory.register(cls.__name__, cls) in AbsModelServer
# __init_subclass__

from uniflow.flow.extract.extract_image_flow import ExtractImageFlow  # noqa: F401, F403
from uniflow.flow.extract.extract_ipynb_flow import ExtractIpynbFlow  # noqa: F401;
from uniflow.flow.extract.extract_md_flow import ExtractMarkdownFlow  # noqa: F401;
from uniflow.flow.extract.extract_pdf_flow import ExtractPDFFlow  # noqa: F401;
from uniflow.flow.extract.extract_txt_flow import ExtractTxtFlow  # noqa: F401, F403

__all__ = [
    "ExtractIpynbFlow",
    "ExtractMarkdownFlow",
    "ExtractPDFFlow",
    "ExtractTxtFlow",
    "ExtractImageFlow",
]



"""Transform markdown op."""

import copy
from dataclasses import dataclass
from typing import Dict, List, Sequence

from uniflow.node import Node
from uniflow.op.op import Op


@dataclass
class HeaderType:
    """Header type."""

    level: int
    name: str
    data: str


@dataclass
class LineType:
    """Line type."""

    metadata: Dict[str, str]
    content: str


class MarkdownHeaderSplitter(Op):
    """Transform markdown class."""

    headers_to_split_on_default = [
        ("#", "Header 1"),
        ("##", "Header 2"),
        ("###", "Header 3"),
        ("####", "Header 4"),
        ("#####", "Header 5"),
        ("######", "Header 6"),
    ]

    def __call__(
        self,
        nodes: Sequence[Node],
        headers_to_split_on_list: List[tuple] = None,
    ) -> Sequence[Node]:
        """Run Model Op.

        Args:
            nodes (Sequence[Node]): Nodes to run.
            headers_to_split_on_list(List[tuple]): Header list for split.

        Returns:
            Sequence[Node]: Nodes after running.
        """
        output_nodes = []
        for node in nodes:
            value_dict = copy.deepcopy(node.value_dict)
            text = value_dict["text"]
            text = self.header_splitter(text.strip(), headers_to_split_on_list)
            output_nodes.append(
                Node(
                    name=self.unique_name(),
                    value_dict={"text": text},
                    prev_nodes=[node],
                )
            )
        return output_nodes

    def header_splitter(
        self,
        markdown_str: str,
        headers_to_split_on_list: List[tuple] = None,
    ):
        """Split markdown by header."""
        if headers_to_split_on_list is None:
            headers_to_split_on_list = self.headers_to_split_on_default

        # Result
        lines_with_metadata: List[LineType] = []

        # Content and metadata of the chunk currently being processed
        current_content: List[str] = []
        current_metadata: Dict[str, str] = {}

        # Keep track of the nested header structure
        header_stack: List[HeaderType] = []
        initial_metadata: Dict[str, str] = {}

        markdown_str += "\n# end"
        lines = markdown_str.split("\n")

        for line in lines:
            stripped_line = line.strip()

            for sep, name in headers_to_split_on_list:
                # Ensure the header as metadata
                if not name:
                    break

                # Check if line starts with a header that we intend to split on
                if stripped_line.startswith(sep) and (
                    # Header with no text OR header is followed by space
                    # Both are valid conditions that sep is being used a header
                    (len(stripped_line) == len(sep) or stripped_line[len(sep)] == " ")
                ):
                    # Get the current header level
                    current_header_level = sep.count("#")

                    # Pop out headers of lower or same level from the stack
                    while header_stack and (
                        header_stack[-1]["level"] >= current_header_level
                    ):
                        # We have encountered a new header
                        # at the same or higher level
                        popped_header = header_stack.pop()
                        # Clear the metadata for the
                        # popped header in initial_metadata
                        if popped_header["name"] in initial_metadata:
                            initial_metadata.pop(popped_header["name"])

                    # Push the current header to the stack
                    sep_length = len(sep)
                    header: HeaderType = {
                        "level": current_header_level,
                        "name": name,
                        "data": stripped_line[sep_length:].strip(),
                    }
                    header_stack.append(header)
                    # Update initial_metadata with the current header
                    initial_metadata[name] = header["data"]

                    # Add the previous line to the lines_with_metadata
                    # only if current_content is not empty
                    if current_content:
                        lines_with_metadata.append(
                            {
                                "content": "\n".join(current_content),
                                "metadata": current_metadata.copy(),
                            }
                        )
                        current_content.clear()

                    break

            if stripped_line:
                current_content.append(stripped_line)
                current_metadata = initial_metadata.copy()

        return [line["content"] for line in lines_with_metadata]



"""Rater __init__ Module."""

# this register all possible model server into ModelServerFactory through
# ModelServerFactory.register(cls.__name__, cls) in AbsModelServer
# __init_subclass__


from uniflow.flow.rater.rater_flow import RaterFlow

__all__ = [
    "RaterFlow",
]



import copy
from typing import Sequence

from uniflow.node import Node
from uniflow.op.op import Op


class PatternSplitter(Op):
    """Pattern Splitter Op Class"""

    def __init__(self, name: str, splitter: str = "\n\n") -> None:
        """Pattern Splitter Op Constructor

        Args:
            name (str): Name of the op.
            splitter (str): Splitter to use. Defaults to "\n\n".
        """
        super().__init__(name)
        self._splitter = splitter

    def __call__(
        self,
        nodes: Sequence[Node],
    ) -> Sequence[Node]:
        """Run Pattern Splitter Op

        Args:
            nodes (Sequence[Node]): Nodes to run.

        Returns:
            Sequence[Node]: Nodes after running the split.
        """
        output_nodes = []
        for node in nodes:
            value_dict = copy.deepcopy(node.value_dict)
            text = value_dict["text"]
            text = text.split(self._splitter)
            output_nodes.append(
                Node(
                    name=self.unique_name(),
                    value_dict={"text": text},
                    prev_nodes=[node],
                )
            )

        return output_nodes



"""Flow class."""
import copy
import logging
from typing import Any, Mapping, Sequence

from uniflow import constants
from uniflow.flow.flow_factory import FlowFactory
from uniflow.node import Node

logger = logging.getLogger(__name__)


class Flow:
    """Flow class."""

    def __init_subclass__(cls) -> None:
        """Initialize subclass."""
        super().__init_subclass__()
        FlowFactory.register(name=cls.__name__, flow_cls=cls)

    def __init__(self) -> None:
        """Initialize Flow class."""
        self._root = None
        logging.basicConfig(format="%(levelname)s [%(module)s]: %(message)s")

    def __call__(self, value_dict: Mapping[str, Any]) -> Mapping[str, Any]:
        """Run flow.

        Args:
            value_dict (Mapping[str, Any]): Input value dict.

        Returns:
            Mapping[str, Any]: Output value dict."""
        nodes = self._enter(value_dict)
        nodes = self.run(nodes)
        output_dict = self._exit(nodes)
        return output_dict

    @property
    def root(self) -> Node:
        """Return root node.

        Returns:
            Node: Root node.
        """
        return self._root

    def _enter(self, value_dict: Mapping[str, Any]) -> Sequence[Node]:
        """Enter flow.

        Args:
            value_dict (Mapping[str, Any]): Input value dict.

        Returns:
            Sequence[Node]: Nodes.
        """
        value_dict = copy.deepcopy(value_dict)
        self._root = Node(name=constants.ROOT_NAME, value_dict=value_dict)
        return [self._root]

    def run(self, nodes: Sequence[Node]) -> Sequence[Node]:
        """Run flow.

        Args:
            nodes (Sequence[Node]): Nodes.

        Returns:
            Sequence[Node]: Nodes.
        """
        return nodes

    def _exit(self, nodes) -> Mapping[str, Any]:
        """Exit flow.

        Args:
            nodes (Sequence[Node]): Nodes.

        Returns:
            Mapping[str, Any]: Output value dict.
        """
        return {
            constants.OUTPUT_NAME: [copy.deepcopy(node.value_dict) for node in nodes],
            constants.ROOT_NAME: self._root,
        }



"""Module for visualizing the results of the computational graph."""

from graphviz import Digraph


class Viz:
    """Visualize the results of the computational graph."""

    @staticmethod
    def to_digraph(self) -> Digraph:
        """Convert the node to graphviz Digraph.

        Returns:
            Digraph: Graphviz Digraph.
        """
        graph = Digraph()
        for node in self.get_all():
            graph.node(node.name)
            for next_node in node.next_nodes:
                graph.edge(node.name, next_node.name)
        return graph
